#!/bin/bash

err(){
    errst=${1:="Unknown error"}
    echo "$0: $errst"
    exit 1
}

#num of args
numargs=$#
compileflag=0
do_compile=0
do_obj=0
skip=0
ofile="a.out"
args=""

CXX=$(which llvm-gcc 2>/dev/null)
[ "$CXX" = "" ] && err "llvm-gcc not found in PATH"

OPT=$(which opt 2>/dev/null)
[ "$OPT" = "" ] && err "opt (the llvm optimizer) not found in PATH"

LLC=$(which llc 2>/dev/null)
[ "$LLC" = "" ] && err "llc (the llvm bitcode compiler) not found in PATH"

# these flags cause opt to avoid reordering accesses to potentially shared memory 
OPTFLAGS="-enable-sc-instcombine -enable-sc-jump-threading -enable-sc-licm -disable-licm-promotion -enable-load-pre=false -enable-sc-gvn -preverify -domtree -verify -lowersetjmp -globalopt -ipsccp -deadargelim -instcombine -simplifycfg -basiccg -prune-eh -inline -functionattrs -domtree -domfrontier -scalarrepl -simplify-libcalls -instcombine -lazy-value-info -jump-threading -correlated-propagation -simplifycfg -instcombine -tailcallelim -simplifycfg -reassociate -domtree -loops -loopsimplify -lcssa -loop-rotate -licm -lcssa -loop-unswitch -instcombine -scalar-evolution -loopsimplify -lcssa -iv-users -indvars -loop-deletion -loop-unroll -instcombine -memdep -gvn -simplifycfg -memdep -sccp -instcombine -lazy-value-info -correlated-propagation -domtree -memdep -adce -simplifycfg -strip-dead-prototypes -print-used-types -deadtypeelim -globaldce -constmerge -preverify -domtree -verify"

# these flags prevent llc from reordering accesses to potentially shared memory
export LLCFLAGS="-combiner-enable-sc -enable-sc-load-chaining -enable-sc-machine-sink -enable-sc-dag-cse"

TMP="$@"
TMP2=${TMP/emit-llvm/}
if (( ${#TMP2} < ${#TMP} )) 
then
  # just use llvm-g++ directly

  ${CXX} $@  

  exit $?
fi

RM_INTERMEDIATE_FILES=0

LLVMFLAGS="-emit-llvm -c "
ENABLE_DEBUG=0
ENABLE_LOG=1

ogiven=0
srcs=""
src=""
count=0
single=1
#flags=${CFLAGS}
#flags="-m32 "
if [ $ENABLE_DEBUG == 1 ] ; then
   flags="-g"
else
   flags=""
fi
code=0


mneedfile=0
mflag=0
margs=""
#following indicates whether -E is implied or not. Not implication of -E, allows compilation, 1 means -E implied
eflag=0
if [ $ENABLE_LOG == 1 ] ; then
   LOG="/tmp/log.txt"
   if [ -e $LOG ] ; then 
      touch $LOG
   fi
fi
#parse args
if [ $ENABLE_LOG == 1 ] ; then
   echo "***CC****************************************************************" >> $LOG
   echo $@ >> $LOG
   echo "*******************************************************************" >> $LOG
fi

for var in $@
do

   skip=0
   code=0

   if [ $ogiven -eq 1 ]
   then
      ogiven=0
      ofile=${var}
      skip=1
   fi

   if [ $mneedfile -eq 1 ]
   then
      margs="${margs} ${var}"
      mneedfile=0
      skip=1
   fi

   if [ ${var} = "-c" ]
   then
      skip=1
      do_compile=1
   fi

   if [ ${var} = "-o" ] 
   then
      ogiven=1
      do_obj=1
      skip=1
   fi

   #handle -M and related flags
   if [[ ${var} = "-M" || ${var} = "-MM" || ${var} = "-MF" || ${var} = "-MG" || ${var} = "-MP" || ${var} = "-MT" || ${var} = "-MQ" || ${var} = "-MD" ||    ${var} = "-MMD" ]] ; then
      mflag=1
      margs="${margs} ${var}"
      skip=1
      if  [[ ${var} = "-MF" ||  ${var} = "-MT" || ${var} = "-MQ" ]] ; then 
         mneedfile=1
      fi

      # -M and -MM has higher priority than -MD or -MMD
      if [[ ${var} = "-M" || ${var} = "-MM" ]] ; then
         eflag=1
      fi

   fi

   if [ ${var:0:2} = "-l" ] ; then
      #linker flag, add this to libs and set skip = 1
      libs="${libs} ${var}"
      if [ $ENABLE_LOG == 1 ] ; then
         echo $var >> $LOG
      fi
      skip=1
   fi

   if [ $ENABLE_LOG == 1 ] ; then
      echo $var >> $LOG
   fi

   if  [[  ${var} =~ [.]cpp$ ||  ${var} =~ [.]c$ || ${var} =~ [.]cc$  || ${var} =~ [.]cxx$ ]] ; then
      skip=1
      srcs="${srcs} ${var}"
      (( count++ ))
      code=1
   fi


   if [ $skip -eq 0 ]
   then
      args="${args} ${var}"
      if [ $code -eq 0 ]
      then
         flags="${flags} ${var}"
      fi
   fi
done

if [ $count -ne 1 ]
then
   single=0
fi

#now executes
if [ $count -eq 0 ] 
then
   #no c or cpp file, better leave everything to gcc directly
   if [ $ENABLE_LOG == 1 ] ; then
      echo " passing to default gcc ${@}" >> $LOG
   fi
   ${CXX} $@
   #gcc $@
  exit $?
else
   for file in ${srcs}
   do
      #basename=${file##*/}
      base=`basename ${file}`
      src=${base%%.*}


      if [ $ENABLE_LOG == 1 ] ; then
         echo $file >> $LOG
         echo "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" >> $LOG
         echo $src >> $LOG
         echo "***********************" >> $LOG
         echo $flags >> $LOG
         echo "##########################" >> $LOG
      fi

      (( count-- ))


      #handle the case of mflag
      if [ $mflag -eq 1 ]
      then
         ${CXX} -M ${margs} $file $flags
      fi

      if [ $do_compile -eq 1 ] && [ $eflag -eq 0 ]
      then
         if [ $ENABLE_LOG == 1 ] ; then
            echo "${CXX} ${LLVMFLAGS} $flags $file -o ${src}.sbc" >> $LOG
         fi
         ${CXX} ${LLVMFLAGS} $flags $file -o "${src}.sbc"
         if [ $? -ne 0 ] ; then
            exit 1
         fi

         if [ $ENABLE_LOG == 1 ] ; then
            echo "${OPT}  ${OPTFLAGS} ${src}.sbc -o ${src}.bc" >> $LOG
         fi
         ${OPT}  ${OPTFLAGS} "${src}.sbc" -o "${src}.bc"

         #llvm-dis "${src}.bc" -o "${src}.ll"
         if [ $ENABLE_LOG == 1 ] ; then
            echo "${LLC}  ${LLCFLAGS} ${src}.bc -o ${src}.s" >> $LOG
         fi
         ${LLC}  ${LLCFLAGS} "${src}.bc" -o "${src}.s"

         if [ ${do_obj} -eq 1 ]
         then
            #this is valid for only one file input for compilation
            if [ $ENABLE_LOG == 1 ] ; then
               echo "${CXX} -c $flags ${src}.s -o ${ofile}" >> $LOG
            fi
            ${CXX} -c $flags "${src}.s" -o ${ofile}
         else
            #single/multiple files for compilation and no -o given
            ${CXX} -c ${flags} "${src}.s" -o "${src}.o"
            if [ $single -eq 1 ] 
            then
               if [ $ENABLE_LOG == 1 ] ; then
                  echo "${CXX} -c ${flags}  ${src}.s"  >> $LOG
               fi
               ${CXX} -c ${flags} "${src}.s" 
            else
               if [ $ENABLE_LOG == 1 ] ; then
                  echo "${CXX} -c ${flags} ${src}.s -o ${src}.o" >> $LOG
               fi
               ${CXX} -c ${flags} "${src}.s" -o "${src}.o"
            fi
         fi

         if [ $RM_INTERMEDIATE_FILES == 1 ]
	     then
	     rm "${src}.sbc" "${src}.bc" "${src}.s"
	 fi
      elif [ $mflag -eq 0 ] && [ $eflag -eq 0 ] ; then
         #full binary
         if [ $ENABLE_LOG == 1 ] ; then
            echo "${CXX} ${LLVMFLAGS} $flags $file -o ${src}.sbc" >> $LOG
         fi
         ${CXX} ${LLVMFLAGS} $flags $file -o "${src}.sbc"
         if [ $? -ne 0 ] ; then
            exit 1
         fi


         if [ $ENABLE_LOG == 1 ] ; then
            echo "${OPT}  ${OPTFLAGS} ${src}.sbc -o ${src}.bc" >> $LOG
         fi
         ${OPT}  ${OPTFLAGS} "${src}.sbc" -o "${src}.bc"

         #llvm-dis "${src}.bc" -o "${src}.ll"
         if [ $ENABLE_LOG == 1 ] ; then
            echo "${LLC}  ${LLCFLAGS} ${src}.bc -o ${src}.s" >> $LOG
         fi
         ${LLC}  ${LLCFLAGS} "${src}.bc" -o "${src}.s"

         #here many files can be given, use counter
         if [ $single -eq 1 ]
         then
            if [ $ENABLE_LOG == 1 ] ; then
               echo "${CXX} ${flags} ${src}.s -o ${ofile} ${libs}" >> $LOG
            fi
            ${CXX} ${flags} "${src}.s" -o ${ofile} ${libs}
         else
            if [ $count -eq 0 ]
            then
               if [ $ENABLE_LOG == 1 ] ; then
                  echo "${CXX} ${flags} *.s -o ${ofile}" >> $LOG
               fi
               ${CXX} ${flags} *".s" -o ${ofile} ${libs}
            fi
         fi
	 
	 if [ $RM_INTERMEDIATE_FILES == 1 ]
	 then
	     rm "${src}.sbc" "${src}.bc" "${src}.s"
	 fi
	 
      fi

      #remove my temp files
   done
fi
